/*
BCM2835 "GPU_FFT" release 3.0
Copyright (c) 2015, Andrew Holme.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the copyright holder nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#define _GNU_SOURCE
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <memory.h>
#include <math.h>
#include <time.h>
#include <getopt.h>

#include "mailbox.h"
#include "gpu_fft.h"

static const char Usage[] =
	"Usage: hello_fft.bin -n log2_N [-j jobs] [-l loops] [-f freq] [-i inc] [-d] [-p]\n"
	"  log2_N = log2(FFT_length),       log2_N = 8...22, <0 = forward transform\n"
	"  jobs   = transforms per batch,   jobs>0,        default 1\n"
	"  loops  = number of test repeats, loops>0,       default 1\n"
	"  freq   = frequency in units of FFT length,      default 1\n"
	"  inc    = frequency increment in case jobs>1,    default 1\n"
	"  -d     = dump result\n"
	"  -p     = print performance counters\n"
	"  -t     = use alternate test pattern\n";

static const char Options[] = "n:j:l:f:i:dpt";

unsigned Microseconds(void)
{
	struct timespec ts;
	clock_gettime(CLOCK_REALTIME, &ts);
	return ts.tv_sec * 1000000 + ts.tv_nsec / 1000;
}

static double chop(double value)
{
	return fabs(value) < 3E-5 ? 0 : value;
}

static double sqr(double d)
{
	return d*d;
}

int main(int argc, char *argv[])
{
	int i, j, k, ret, loops, fbias, finc, freq, log2_N, jobs,
		debug, perf, test, N, dir, bad, mb = mbox_open();
	unsigned t[3];
	double tsq[4];

	struct GPU_FFT_COMPLEX *base;
	struct GPU_FFT *fft;

	log2_N = 0;
	jobs = 1;
	loops = 1;
	fbias = 1;
	finc = 1;
	debug = 0;
	perf = 0;
	test = 0;

	if (argc < 2)
	{usage:
		printf(Usage);
		return -1;
	}

	j = 0;
 next:
	if (j)
		i = -1;
	else
		i = getopt(argc, argv, Options);
 oldopt:
	//printf("i=%c, optarg=%s\n", i, optarg);
	switch (i)
	{default:
		goto usage;
	 case 'n':
		log2_N = atoi(optarg);
		goto next;
	 case 'l':
		loops = atoi(optarg);
		goto next;
	 case 'j':
		jobs = atoi(optarg);
		goto next;
	 case 'f':
		fbias = atoi(optarg);
		goto next;
	 case 'i':
		finc = atoi(optarg);
		goto next;
	 case 'd':
		debug = 1;
		goto next;
	 case 'p':
		perf = 1;
		goto next;
	 case 't':
		test = 1;
 		goto next;
	 case -1: // end: old style syntax?
		if (optind >= argc)
			break;
		if (j >= 10)
			goto usage;
		i = Options[j];
		optarg = argv[optind++];
		j += 2;
		goto oldopt;
	}

	dir = GPU_FFT_REV;
	if (log2_N < 0)
	{
		dir = GPU_FFT_FWD;
		log2_N = -log2_N;
	}
	if (jobs < 1 || loops < 1)
	{
		fprintf(stderr, Usage);
		return -1;
	}

	N = 1 << log2_N; // FFT length
	ret = gpu_fft_prepare(mb, log2_N, dir, jobs, &fft); // call once
	if (ret) return ret;

	// setup performance counters
	if (perf)
		gpu_fft_pct_setup(&fft->base,
			1<<13 | 1<<14 | 1<<16 | 1<<17 | 1<<20 | 1<<21 | 1<<22 | 1<<23 | 1<<24 | 1<<25 | 1<<28 | 1<<29);

	switch (ret)
	{
	case -1:
		fprintf(stderr, "Unable to enable V3D. Please check your firmware is up to date.\n");
		return -1;
	case -2:
		fprintf(stderr, "log2_N=%d not supported.  Try between 8 and 22.\n", log2_N);
		return -1;
	case -3:
		fprintf(stderr, "Out of memory.  Try a smaller batch or increase GPU memory.\n");
		return -1;
	case -4:
		fprintf(stderr, "Unable to map Videocore peripherals into ARM memory space.\n");
		return -1;
	case -5:
		fprintf(stderr, "Can't open libbcm_host.\n");
		return -1;
	case -6:
		fprintf(stderr, "Can't open obj/twiddles.bin.\n");
		return -1;
	}

	t[2] = 0;
	for (k = 0; k < loops; k++)
	{
		for (j = 0; j < jobs; j++)
			memset(fft->out + j * fft->step, -1, N * sizeof(struct GPU_FFT_COMPLEX));

		for (j = 0; j < jobs; j++)
		{
			base = fft->in + j * fft->step; // input buffer
			if (dir == GPU_FFT_REV)
			{
				freq = j * finc + fbias & (N - 1);
				memset(base, 0, N * sizeof(struct GPU_FFT_COMPLEX));
				base[freq].re = 0.5;
				base[N - freq & N - 1].re += 0.5;
				if (test)
				{	base[freq].re += base[freq].im = .5 * M_SQRT1_2;
					base[N - freq & N - 1].re -= .5 * M_SQRT1_2;
					base[N - freq & N - 1].im += .5 * M_SQRT1_2;
				}
			} else
			{
				double s, c;
				freq = j * finc + fbias & (N - 1);
				for (i = 0; i < N; i++)
				{	sincos((freq * i & (N - 1)) * M_2PI / N, &s, &c);
					base[i].re = c;
					base[i].im = s;
				}
				if (test)
				{	++freq;
					for (i = 0; i < N; i++)
					{	sincos((freq * i + (N>>3) & (N - 1)) * M_2PI / N, &s, &c);
						base[i].re += c;
						base[i].im += s;
					}
				}
			}
		}

		usleep(1); // Yield to OS
		t[0] = Microseconds();
		gpu_fft_execute(fft); // call one or many times
		t[1] = Microseconds();
		if (k)
			t[2] += t[1] - t[0];

		if (perf)
		{	unsigned counters[16] = {0};
			i = gpu_fft_pct_read(&fft->base, counters);
			printf(
				"GPU load\t%5.2f %%\n"
				"GPU exec instr\t%5.2f %%\n"
				"TMU stall\t%5.2f %%\n"
				"Inst cache miss\t%5.2f %%\t%7u\n"
				"Unif cache miss\t%5.2f %%\t%7u\n"
				"TMU cache miss\t%5.2f %%\t%7u\n"
				"L2C cache miss\t%5.2f %%\t%7u\n",
				100.*counters[1]/(counters[0]+counters[1]),
				100.*counters[2]/counters[1],
				100.*counters[3]/counters[1],
				100.*counters[5]/(counters[4]+counters[5]), counters[5],
				100.*counters[7]/(counters[6]+counters[7]), counters[7],
				100.*counters[9]/counters[8], counters[9],
				100.*counters[11]/(counters[10]+counters[11]), counters[11] );
		}

		tsq[2] = tsq[3] = 0;
		bad = 0;
		for (j = 0; j < jobs; j++)
		{
			struct GPU_FFT_COMPLEX *base2;
			tsq[0] = tsq[1] = 0;
			base = fft->out + j * fft->step; // output buffer
			base2 = fft->in + j * fft->step; // output buffer
			if (debug)
				puts("i\tf\t    out.re\t    out.im\t   out'.re\t   out'.im\t      chi2\t     in.re\t     in.im");
			if (dir == GPU_FFT_REV)
			{
				freq = j * finc + fbias & (N - 1);
				tsq[0] += N/2 * (test + 1);
				for (i = 0; i < N; i++)
				{
					double s, c, r;
					c = cos((freq * i & (N - 1)) * M_2PI / N);
					if (!test)
					{	r = sqr(c - base[i].re) + sqr(base[i].im);
						s = 0;
					} else
					{	s = sin(((freq * i + (N>>3)) & (N - 1)) * M_2PI / N);
						r = sqr(c - base[i].re) + sqr(s - base[i].im);
					}
					tsq[1] += r;
					if (debug)
						printf("%d\t%d\t%10g\t%10g\t%10g\t%10g\t%10g\t%10g\t%10g\n", i, freq, chop(base[i].re), chop(base[i].im), chop(c), chop(s), chop(r), chop(base2[i].re), chop(base2[i].im));
				}
			} else
			{
				freq = j * finc + fbias & (N - 1);
				tsq[0] += 2 * N;
				for (i = 0; i < N; i++)
				{
					double re = 0, im = 0;
					if (i == freq)
						re = N;
					else if (test && i == ((freq + 1) & (N - 1)))
						im = re = M_SQRT1_2 * N;
					double r = sqr(re - base[i].re) + sqr(im - base[i].im);
					tsq[1] += r;
					if (debug)
						printf("%d\t%d\t%10g\t%10g\t%10g\t%10g\t%10g\t%10g\t%10g\n", i, freq, chop(base[i].re), chop(base[i].im), chop(re), chop(im), chop(r), chop(fft->in[i].re), chop(fft->in[i].im));
				}
			}
			//fprintf(stderr, "step_rms_err = %.5e, j = %d\n", sqrt(tsq[1] / tsq[0]), j);
			tsq[2] += tsq[0];
			tsq[3] += tsq[1];
			if (tsq[1] / tsq[0] > 2E-6)
				++bad;
		}

		fprintf(stderr, "rel_rms_err = %.5e, usecs = %f, k = %d\n", sqrt(tsq[3] / tsq[2]), (double)(t[1] - t[0]) / jobs, k);
		if (bad)
			fprintf(stderr, "failed: %i = %f %%\n", bad, 100. * bad / jobs);
	}

	gpu_fft_release(fft); // Videocore memory lost if not freed !

	if (loops > 1)
		fprintf(stderr, "average: usecs = %f\n", (double)t[2] / jobs / (loops - 1));
	return 0;
}
